#include "plugin-djvu.h"

#define SCALE_FACTOR 0.2

iDjvu::iDjvu(void)
	:d_context(0)
	,d_document(0)
	,d_format(0)
	,d_page_width(0)
	,d_page_height(0)
	,d_page_scale(1.0)
{
	g_debug("Creating iDjvu");
	d_context = ddjvu_context_create ("Gloobus-Preview");
	d_format = ddjvu_format_create (DDJVU_FORMAT_RGB24, 0, 0);
	ddjvu_format_set_row_order (d_format, 1);
}

iDjvu::~iDjvu(void)
{
	g_debug("Destroying iDjvu");
	if(d_context)
		ddjvu_context_release(d_context);
	if(d_document)
		ddjvu_document_release(d_document);
	if(d_format)
		ddjvu_format_release(d_format);
}

void iDjvu::end_plugin(void)
{
	g_debug("Ending iDjvu");
}
//====================== LOAD ==============================//
bool iDjvu::load_doc()
{
	gchar *filename;
	GError **error = NULL;
	GError *djvu_error = NULL;
	
	filename = g_filename_from_uri (g_file_get_uri(m_gfile), NULL, error);
	if (!filename) {
		set_error( g_error_new(g_quark_from_string ("iDjvu"), 301,
			_("Error getting filename")) );
		if(*error)
			g_error_free(*error);
		return false;
	}
	
	d_document = ddjvu_document_create_by_filename (d_context, filename, TRUE);
	if (!d_document) {
		set_error( g_error_new(g_quark_from_string ("iDjvu"), 302,
			_("Document has incorrect format")) );
		g_free (filename);
		return false;
	}
	
	wait_for_message (DDJVU_DOCINFO, &djvu_error);
	if (djvu_error) {
		set_error( g_error_new(g_quark_from_string ("iDjvu"), 303,
			"%s:\n%s", _("An error occurred"), djvu_error->message) );
		g_free (filename);
		g_error_free (djvu_error);
		return false;
	}

	bool missing_files = false;
	if (ddjvu_document_get_type (d_document) == DDJVU_DOCTYPE_INDIRECT) {
		gint n_files;
		gint i;
		gchar *base;

		base = g_path_get_dirname (filename);

		n_files = ddjvu_document_get_filenum (d_document);
		for (i = 0; i < n_files; i++) {
			struct ddjvu_fileinfo_s fileinfo;
			gchar *file;
			
			ddjvu_document_get_fileinfo (d_document, i, &fileinfo);

			if (fileinfo.type != 'P')
				continue;

			file = g_build_filename (base, fileinfo.id, NULL);
			if (!g_file_test (file, G_FILE_TEST_EXISTS)) {
				missing_files = TRUE;
				g_free (file);
				
				break;
			}
			g_free (file);
		}
		g_free (base);
	}
	g_free (filename);

	if (missing_files) {
		set_error( g_error_new(g_quark_from_string ("iDjvu"), 304,
			_("The document is composed of several files. One or more of such files cannot be accessed")) );
		return false;
	}
	get_page_scale();

	return true;
}

int iDjvu::get_n_pages()
{
	return ddjvu_document_get_pagenum (d_document);
}

void iDjvu::get_page_size (gint page, int *width, int *height)
{
	ddjvu_pageinfo_t info;
	ddjvu_status_t r;
	
	while ((r = ddjvu_document_get_pageinfo(d_document, page, &info)) < DDJVU_JOB_OK)
		handle_events(TRUE, NULL);
	
	if (r >= DDJVU_JOB_FAILED)
		handle_events(TRUE, NULL);
	//*width = info.width * SCALE_FACTOR;
	*width = info.width;
	//*height = info.height * SCALE_FACTOR;
	*height = info.height;
}


//====================== GET PAGES WIDTH =======================//
int iDjvu::get_page_width_unscaled  ( void )
{
	return d_page_width;
}
//====================== GET PAGES HEIGHT =======================//
int	iDjvu::get_page_height_unscaled ( void )
{
	return d_page_height;
}
//====================== GET PAGE PIXBUF =======================//
GdkPixbuf* iDjvu::get_page_pixbuf ( int n_page )
{
	GdkPixbuf* pixbuf;
	gchar *pixels;
	gint   rowstride;
	ddjvu_rect_t rrect;
	ddjvu_rect_t prect;
	ddjvu_page_t *d_page;
	int width;
	int height;

	d_page = ddjvu_page_create_by_pageno (d_document, n_page);
	while (!ddjvu_page_decoding_done (d_page))
		handle_events(TRUE, NULL);

	width = (int)(ddjvu_page_get_width (d_page) * d_page_scale);
	height = (int)(ddjvu_page_get_height (d_page) * d_page_scale);

	pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, false, 8,
		(int)width, (int)height);

	rowstride = gdk_pixbuf_get_rowstride (pixbuf);
	pixels = (gchar*) gdk_pixbuf_get_pixels (pixbuf);
	prect.x = 0;
	prect.y = 0;
	prect.w = width;
	prect.h = height;
	rrect = prect;

	ddjvu_page_render (d_page, DDJVU_RENDER_COLOR, &prect, &rrect, d_format,
		rowstride, pixels);

	ddjvu_page_release(d_page);

	return pixbuf;
	
}

void iDjvu::get_page_scale ( void )
{
	get_page_size (0, &d_page_width, &d_page_height);
	
	int sw = get_screen_width() - 150;

	if(d_page_width <= sw) {
		d_page_scale = 1;
	} else {
		d_page_scale = (double)sw / (double)d_page_width;
	}
}

int iDjvu::get_page_width( void )
{
	return (int)(d_page_width * d_page_scale);
}

int iDjvu::get_page_height( void )
{
	return (int)(d_page_height * d_page_scale);
}

GtkWidget* iDjvu::get_page(int n) {
	GdkPixbuf * temp = get_page_pixbuf(n);
	GtkWidget * page = gtk_image_new_from_pixbuf(temp);
	g_object_unref(temp);
	return page;
}

void iDjvu::wait_for_message (ddjvu_message_tag_t message, GError **error)
{
	const ddjvu_message_t *msg;

	ddjvu_message_wait (d_context);
	while ((msg = ddjvu_message_peek (d_context)) && (msg->m_any.tag != message)) {
		handle_message (msg, error);
		ddjvu_message_pop (d_context);
		if (error && *error)
			return;
	}
	if (msg && msg->m_any.tag == message)
		ddjvu_message_pop (d_context);
}

void iDjvu::handle_message (const ddjvu_message_t *msg, GError **error)
{
	switch (msg->m_any.tag) {
		case DDJVU_ERROR: {
			gchar *error_str;
			
			if (msg->m_error.filename) {
				error_str = g_strdup_printf ("DjvuLibre error: %s:%d",
							     msg->m_error.filename,
							     msg->m_error.lineno);
			} else {
				error_str = g_strdup_printf ("DjvuLibre error: %s",
							     msg->m_error.message);
			}
			
			//if (error) {
			//	g_set_error_literal (error, EV_DJVU_ERROR, 0, error_str);
			//} else {
				g_warning ("%s", error_str);
			//}
				
			g_free (error_str);
			return;
		}
			break;
		default:
			break;
	}
}

void iDjvu::handle_events (int wait, GError **error)
{
	const ddjvu_message_t *msg;
	
	if (!d_context)
		return;

	if (wait)
		ddjvu_message_wait (d_context);

	while ((msg = ddjvu_message_peek (d_context))) {
		handle_message (msg, error);
		ddjvu_message_pop (d_context);
		if (error && *error)
			return;
	}
}
